@@ -248,6 +248,7 @@ print_usage(const char *prog_name)
            "  -l,--lines                     number of matches to show\n"
            "  -w,--width                     window width, in characters (margins and\n"
            "                                 borders not included)\n"
+           "     --tabs=INT                  number of spaces a tab is expanded to (8)\n"
            "  -x,--horizontal-pad=PAD        horizontal padding, in pixels (40)\n"
            "  -y,--vertical-pad=PAD          vertical padding, in pixels (8)\n"
            "  -P,--inner-pad=PAD             vertical padding between prompt and match list,\n"
@@ -278,6 +279,8 @@ print_usage(const char *prog_name)
            "     --no-exit-on-keyboard-focus-loss  do not exit when losing keyboard focus\n"
            "     --launch-prefix=COMMAND     prefix to add before argv of executed program\n"
            "  -d,--dmenu                     dmenu compatibility mode\n"
+           "     --dmenu0                    like --dmenu, but input is NUL separated\n"
+           "                                 instead of newline separated\n"
            "     --index                     print selected entry's index instead of of the \n"
            "                                 entry's text (dmenu mode only)\n"
            "  -R,--no-run-if-empty           exit immediately without showing UI if stdin\n"
@@ -409,18 +412,38 @@ populate_apps(void *_ctx)
 {
     struct context *ctx = _ctx;
     struct application_list *apps = ctx->apps;
-    const char *icon_theme = ctx->conf->icon_theme;
-    const char *terminal = ctx->conf->terminal;
-    bool actions_enabled = ctx->conf->actions_enabled;
-    bool dmenu_enabled = ctx->conf->dmenu.enabled;
-    bool icons_enabled = ctx->conf->icons_enabled;
+    const struct config *conf = ctx->conf;
+    const char *icon_theme = conf->icon_theme;
+    const char *terminal = conf->terminal;
+    bool actions_enabled = conf->actions_enabled;
+    bool dmenu_enabled = conf->dmenu.enabled;
+    bool icons_enabled = conf->icons_enabled;
+    char dmenu_delim = conf->dmenu.delim;
+    bool filter_desktop = conf->filter_desktop;
+    char_list_t desktops = tll_init();
+    char *saveptr = NULL;
+
+    if (filter_desktop) {
+        char *xdg_current_desktop = getenv("XDG_CURRENT_DESKTOP");
+        if (xdg_current_desktop && strlen(xdg_current_desktop) != 0) {
+            xdg_current_desktop = strdup(xdg_current_desktop);
+            for (char *desktop = strtok_r(xdg_current_desktop, ":", &saveptr);
+                 desktop != NULL;
+                 desktop = strtok_r(NULL, ":", &saveptr))
+                {
+                    tll_push_back(desktops, strdup(desktop));
+                }
+            free(xdg_current_desktop);
+        }
+    }
 
     if (dmenu_enabled)
-        dmenu_load_entries(apps, ctx->dmenu_abort_fd);
+        dmenu_load_entries(apps, dmenu_delim, ctx->dmenu_abort_fd);
     else {
-        xdg_find_programs(terminal, actions_enabled, apps);
+        xdg_find_programs(terminal, actions_enabled, filter_desktop, &desktops, apps);
         read_cache(apps);
     }
+    tll_free_and_free(desktops, free);
 
     int r = send_event(ctx->event_fd, EVENT_APPS_LOADED);
     if (r != 0)
@@ -515,6 +538,9 @@ main(int argc, char *const *argv)
     #define OPT_LAYER                        269
     #define OPT_ICON_THEME                   270
     #define OPT_NO_EXIT_ON_KB_LOSS           271
+    #define OPT_TABS                         272
+    #define OPT_DMENU_NULL                   273
+    #define OPT_FILTER_DESKTOP               274
 
     static const struct option longopts[] = {
         {"config",               required_argument, 0,  OPT_CONFIG},
@@ -527,6 +553,7 @@ main(int argc, char *const *argv)
         {"password",             optional_argument, 0, OPT_PASSWORD},
         {"lines",                required_argument, 0, 'l'},
         {"width",                required_argument, 0, 'w'},
+        {"tabs",                 required_argument, 0, OPT_TABS},
         {"horizontal-pad",       required_argument, 0, 'x'},
         {"vertical-pad",         required_argument, 0, 'y'},
         {"inner-pad",            required_argument, 0, 'P'},
@@ -542,6 +569,7 @@ main(int argc, char *const *argv)
         {"prompt",               required_argument, 0, 'p'},
         {"terminal",             required_argument, 0, 'T'},
         {"show-actions",         no_argument,       0, OPT_SHOW_ACTIONS},
+        {"filter-desktop",       optional_argument, 0, OPT_FILTER_DESKTOP},
         {"no-fuzzy",             no_argument,       0, OPT_NO_FUZZY},
         {"fuzzy-min-length",     required_argument, 0, OPT_FUZZY_MIN_LENGTH},
         {"fuzzy-max-length-discrepancy", required_argument, 0, OPT_FUZZY_MAX_LENGTH_DISCREPANCY},
@@ -554,6 +582,7 @@ main(int argc, char *const *argv)
 
         /* dmenu mode */
         {"dmenu",                no_argument,       0, 'd'},
+        {"dmenu0",               no_argument,       0, OPT_DMENU_NULL},
         {"no-run-if-empty",      no_argument,       0, 'R'},
         {"index",                no_argument,       0, OPT_DMENU_INDEX},
 
@@ -579,6 +608,7 @@ main(int argc, char *const *argv)
         bool icons_disabled_set:1;
         bool lines_set:1;
         bool chars_set:1;
+        bool tabs_set:1;
         bool pad_x_set:1;
         bool pad_y_set:1;
         bool pad_inner_set:1;
@@ -592,6 +622,7 @@ main(int argc, char *const *argv)
         bool border_size_set:1;
         bool border_radius_set:1;
         bool actions_enabled_set:1;
+        bool filter_desktop_set:1;
         bool fuzzy_set:1;
         bool fuzzy_min_length_set:1;
         bool fuzzy_max_length_discrepancy_set:1;
@@ -601,11 +632,13 @@ main(int argc, char *const *argv)
         bool dmenu_enabled_set:1;
         bool dmenu_mode_set:1;
         bool dmenu_exit_immediately_if_empty_set:1;
+        bool dmenu_delim_set:1;
         bool layer_set:1;
         bool no_exit_on_keyboard_focus_loss_set:1;
     } cmdline_overrides = {{0}};
 
     setlocale(LC_CTYPE, "");
+    setlocale(LC_MESSAGES, "");
 
     /* Auto-enable dmenu mode if invoked through a ‘dmenu’ symlink */
     if (argv[0] != NULL) {
@@ -724,7 +757,7 @@ main(int argc, char *const *argv)
         }
 
         case OPT_PASSWORD: {
-            char32_t password_char = U'*';
+            char32_t password_char = U'\0';
             if (optarg != NULL) {
                 char32_t *wide_optarg = ambstoc32(optarg);
                 if (c32len(wide_optarg) != 1) {
@@ -738,7 +771,8 @@ main(int argc, char *const *argv)
                 password_char = wide_optarg[0];
                 free(wide_optarg);
             }
-            cmdline_overrides.conf.password = password_char;
+            cmdline_overrides.conf.password_mode.enabled = true;
+            cmdline_overrides.conf.password_mode.character = password_char;
             break;
         }
 
@@ -762,6 +796,14 @@ main(int argc, char *const *argv)
             cmdline_overrides.chars_set = true;
             break;
 
+        case OPT_TABS:
+            if (sscanf(optarg, "%u", &cmdline_overrides.conf.tabs) != 1) {
+                fprintf(stderr, "%s: invalid tab count\n", optarg);
+                return EXIT_FAILURE;
+            }
+            cmdline_overrides.tabs_set = true;
+            break;
+
         case 'x':
             if (sscanf(optarg, "%u", &cmdline_overrides.conf.pad.x) != 1) {
                 fprintf(stderr, "%s: invalid padding\n", optarg);
@@ -802,9 +844,12 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t background;
-            if (sscanf(clr_start, "%08x", &background) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t background = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "background-color: %s: invalid color\n",
+                        optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.background =
@@ -818,9 +863,11 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t text_color;
-            if (sscanf(clr_start, "%08x", &text_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t text_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "text-color: %s: invalid color\n", optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.text = conf_hex_to_rgba(text_color);
@@ -833,9 +880,11 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t match_color;
-            if (sscanf(clr_start, "%x", &match_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t match_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "match-color: %s: invalid color\n", optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.match =
@@ -849,9 +898,11 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t selection_color;
-            if (sscanf(clr_start, "%x", &selection_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t selection_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "selection-color: %s: invalid color\n", optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.selection =
@@ -865,9 +916,12 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t selection_text_color;
-            if (sscanf(clr_start, "%x", &selection_text_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t selection_text_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "selection-text-color: %s: invalid color\n",
+                        optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.selection_text =
@@ -881,9 +935,12 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t selection_match_color;
-            if (sscanf(clr_start, "%x", &selection_match_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t selection_match_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "selection-match-color: %s: invalid color\n",
+                        optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.selection_match =
@@ -916,9 +973,11 @@ main(int argc, char *const *argv)
             if (clr_start[0] == '#')
                 clr_start++;
 
-            uint32_t border_color;
-            if (sscanf(clr_start, "%x", &border_color) != 1) {
-                fprintf(stderr, "%s: invalid color\n", optarg);
+            errno = 0;
+            char *end = NULL;
+            uint32_t border_color = strtoul(clr_start, &end, 16);
+            if (errno != 0 || end == NULL || *end != '\0' || (end - clr_start) != 8) {
+                fprintf(stderr, "border-color: %s: invalid color\n", optarg);
                 return EXIT_FAILURE;
             }
             cmdline_overrides.conf.colors.border =
@@ -932,6 +991,18 @@ main(int argc, char *const *argv)
             cmdline_overrides.conf.actions_enabled = true;
             break;
 
+        case OPT_FILTER_DESKTOP:
+            cmdline_overrides.filter_desktop_set = true;
+            if (optarg != NULL && strcasecmp(optarg, "no") == 0)
+                cmdline_overrides.conf.filter_desktop = false;
+            else if (optarg != NULL) {
+                fprintf(stderr, "%s: invalid filter-desktop option\n", optarg);
+                return EXIT_FAILURE;
+            }
+            else
+                cmdline_overrides.conf.filter_desktop = true;
+            break;
+
         case OPT_NO_FUZZY:
             cmdline_overrides.conf.fuzzy.enabled = false;
             cmdline_overrides.fuzzy_set = true;
@@ -1014,6 +1085,13 @@ main(int argc, char *const *argv)
             cmdline_overrides.dmenu_enabled_set = true;
             break;
 
+        case OPT_DMENU_NULL:
+            cmdline_overrides.conf.dmenu.enabled = true;
+            cmdline_overrides.conf.dmenu.delim = '\0';
+            cmdline_overrides.dmenu_enabled_set = true;
+            cmdline_overrides.dmenu_delim_set = true;
+            break;
+
         case 'R':
             cmdline_overrides.conf.dmenu.exit_immediately_if_empty = true;
             cmdline_overrides.dmenu_exit_immediately_if_empty_set = true;
@@ -1063,11 +1141,11 @@ main(int argc, char *const *argv)
             return EXIT_SUCCESS;
 
         case ':':
-            fprintf(stderr, "error: -%c: missing required argument\n", optopt);
+            fprintf(stderr, "error: %s: missing required argument\n", argv[optind-1]);
             return EXIT_FAILURE;
 
         case '?':
-            fprintf(stderr, "error: -%c: invalid option\n", optopt);
+            fprintf(stderr, "error: %s: invalid option\n", argv[optind-1]);
             return EXIT_FAILURE;
         }
     }
@@ -1093,8 +1171,15 @@ main(int argc, char *const *argv)
         free(conf.prompt);
         conf.prompt = cmdline_overrides.conf.prompt;
     }
-    if (cmdline_overrides.conf.password != U'\0')
-        conf.password = cmdline_overrides.conf.password;
+    if (cmdline_overrides.conf.password_mode.enabled) {
+        conf.password_mode.enabled = true;
+        conf.password_mode.character =
+            cmdline_overrides.conf.password_mode.character != U'\0'
+            ? cmdline_overrides.conf.password_mode.character
+            : conf.password_mode.character != U'\0'
+            ? conf.password_mode.character
+            : U'*';
+    }
     if (cmdline_overrides.conf.terminal != NULL) {
         free(conf.terminal);
         conf.terminal = strdup(cmdline_overrides.conf.terminal);
@@ -1121,6 +1206,8 @@ main(int argc, char *const *argv)
         conf.lines = cmdline_overrides.conf.lines;
     if (cmdline_overrides.chars_set)
         conf.chars = cmdline_overrides.conf.chars;
+    if (cmdline_overrides.tabs_set)
+        conf.tabs = cmdline_overrides.conf.tabs;
     if (cmdline_overrides.pad_x_set)
         conf.pad.x = cmdline_overrides.conf.pad.x;
     if (cmdline_overrides.pad_y_set)
@@ -1152,6 +1239,8 @@ main(int argc, char *const *argv)
                  cmdline_overrides.conf.border.radius);
 #endif
     }
+    if (cmdline_overrides.filter_desktop_set)
+        conf.filter_desktop = cmdline_overrides.conf.filter_desktop;
     if (cmdline_overrides.actions_enabled_set)
         conf.actions_enabled = cmdline_overrides.conf.actions_enabled;
     if (cmdline_overrides.fuzzy_set)
@@ -1172,6 +1261,8 @@ main(int argc, char *const *argv)
         conf.exit_on_kb_focus_loss = cmdline_overrides.conf.exit_on_kb_focus_loss;
     if (cmdline_overrides.dmenu_enabled_set)
         conf.dmenu.enabled = cmdline_overrides.conf.dmenu.enabled;
+    if (cmdline_overrides.dmenu_delim_set)
+        conf.dmenu.delim = cmdline_overrides.conf.dmenu.delim;
     if (cmdline_overrides.dmenu_mode_set)
         conf.dmenu.mode = cmdline_overrides.conf.dmenu.mode;
     if (cmdline_overrides.dmenu_exit_immediately_if_empty_set)
@@ -1248,7 +1339,7 @@ main(int argc, char *const *argv)
              * If no_run_if_empty is set, we *must* load the entries
              * *before displaying the window.
              */
-            dmenu_load_entries(apps, -1);
+            dmenu_load_entries(apps, conf.dmenu.delim, -1);
             if (apps->count == 0)
                 goto out;
 
