@@ -1,11 +1,13 @@
 #include "xdg.h"
 
+#include <errno.h>
+#include <limits.h>
+#include <locale.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdbool.h>
 #include <unistd.h>
-#include <limits.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -21,9 +23,66 @@
 
 typedef tll(struct application) application_llist_t;
 
+struct locale_variants {
+    char *lang_country_modifier;
+    char *lang_country;
+    char *lang_modifier;
+    char *lang;
+};
+
+struct action {
+    char32_t *name;
+    char32_t *generic_name;
+    char *app_id;
+    char32_t *comment;
+    char32_list_t keywords;
+    char32_list_t categories;
+    char_list_t onlyshowin;
+    char_list_t notshowin;
+
+    int name_locale_score;
+    int generic_name_locale_score;
+    int comment_locale_score;
+    int keywords_locale_score;
+    int categories_locale_score;
+
+    char *icon;
+    char *exec;
+    char32_t *wexec;
+
+    char *path;
+    bool visible;
+    bool use_terminal;
+};
+
+static bool
+filter_desktop_entry(const struct action *act, const char_list_t *desktops)
+{
+    /* If a matching entry is found in OnlyShowIn then the desktop file is
+     * shown. If an entry is found in NotShowIn then the desktop file is not
+     * shown. */
+    tll_foreach(*desktops, current) {
+        tll_foreach(act->onlyshowin, desktop) {
+            if (strcmp(current->item, desktop->item) == 0)
+                return true;
+        }
+
+        tll_foreach(act->notshowin, desktop) {
+            if (strcmp(current->item, desktop->item) == 0)
+                return false;
+        }
+    }
+
+    /* By default, a desktop file should be shown, unless an OnlyShowIn key is
+     * present, in which case, the default is for the file not to be shown. */
+    return tll_length(act->onlyshowin) == 0;
+}
+
 static void
 parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                    const char *terminal, bool include_actions,
+                   bool filter_desktops, char_list_t *desktops,
+                   const struct locale_variants *lc_messages,
                    application_llist_t *applications)
 {
     FILE *f = fdopen(fd, "r");
@@ -34,22 +93,6 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
 
     tll(char *) action_names = tll_init();
 
-    struct action {
-        char32_t *name;
-        char32_t *generic_name;
-        char32_t *comment;
-        char32_list_t keywords;
-        char32_list_t categories;
-
-        char *icon;
-        char *exec;
-        char32_t *wexec;
-
-        char *path;
-        bool visible;
-        bool use_terminal;
-    };
-
     tll(struct action) actions = tll_init();
     tll_push_back(actions, ((struct action){.visible = true}));
     struct action *action = &tll_back(actions);
@@ -136,13 +179,54 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
             }
         }
 
-        const char *key = strtok(line, "=");
-        char *value = strtok(NULL, "\n");
+        char *ctx;
+        char *key = strtok_r(line, "=", &ctx);
+        char *value = strtok_r(NULL, "\n", &ctx);
+        int locale_score = 1;  /* Default, locale not specified in key */
 
         if (key != NULL && value != NULL) {
+            const size_t key_len = strlen(key);
+
+            if (key[key_len - 1] == ']') {
+                char *locale = strchr(key, '[');
+                if (locale != NULL) {
+                    /* NULL terminate key */
+                    *locale = '\0';
+
+                    /* NULL terminate locale */
+                    key[key_len - 1] = '\0';
+                    locale++;  /* Skip past ‘[’ */
+
+                    if (lc_messages->lang_country_modifier != NULL &&
+                        strcmp(locale, lc_messages->lang_country_modifier) == 0)
+                    {
+                        locale_score = 5;
+                    } else if (lc_messages->lang_country != NULL &&
+                               strcmp(locale, lc_messages->lang_country) == 0)
+                    {
+                        locale_score = 4;
+                    } else if (lc_messages->lang_modifier != NULL &&
+                               strcmp(locale, lc_messages->lang_modifier) == 0)
+                    {
+                        locale_score = 3;
+                    } else if (lc_messages->lang != NULL &&
+                               strcmp(locale, lc_messages->lang) == 0)
+                    {
+                        locale_score = 2;
+                    } else {
+                        /* Key has locale, but didn’t match. Make sure
+                         * we don’t use its value */
+                        locale_score = -1;
+                    }
+                }
+            }
+
             if (strcmp(key, "Name") == 0) {
-                free(action->name);
-                action->name = ambstoc32(value);
+                if (locale_score > action->name_locale_score) {
+                    free(action->name);
+                    action->name = ambstoc32(value);
+                    action->name_locale_score = locale_score;
+                }
             }
 
             else if (strcmp(key, "Exec") == 0) {
@@ -158,46 +242,83 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
             }
 
             else if (strcmp(key, "GenericName") == 0) {
-                free(action->generic_name);
-                action->generic_name = ambstoc32(value);
+                if (locale_score > action->generic_name_locale_score) {
+                    free(action->generic_name);
+                    action->generic_name = ambstoc32(value);
+                    action->generic_name_locale_score = locale_score;
+                }
+            }
+
+            else if (strcmp(key, "StartupWMClass") == 0) {
+                free(action->app_id);
+                action->app_id = strdup(value);
             }
 
             else if (strcmp(key, "Comment") == 0) {
-                free(action->comment);
-                action->comment = ambstoc32(value);
+                if (locale_score > action->comment_locale_score) {
+                    free(action->comment);
+                    action->comment = ambstoc32(value);
+                    action->comment_locale_score = locale_score;
+                }
             }
 
             else if (strcmp(key, "Keywords") == 0) {
-                for (const char *kw = strtok(value, ";");
-                     kw != NULL;
-                     kw = strtok(NULL, ";"))
-                {
-                    char32_t *wide_kw = ambstoc32(kw);
-                    if (wide_kw != NULL)
-                        tll_push_back(action->keywords, wide_kw);
+                if (locale_score > action->keywords_locale_score) {
+                    for (const char *kw = strtok_r(value, ";", &ctx);
+                         kw != NULL;
+                         kw = strtok_r(NULL, ";", &ctx))
+                    {
+                        char32_t *wide_kw = ambstoc32(kw);
+                        if (wide_kw != NULL)
+                            tll_push_back(action->keywords, wide_kw);
+                    }
+
+                    action->keywords_locale_score = locale_score;
                 }
             }
 
             else if (strcmp(key, "Categories") == 0) {
-                for (const char *category = strtok(value, ";");
-                     category != NULL;
-                     category = strtok(NULL, ";"))
-                {
-                    char32_t *wide_category = ambstoc32(category);
-                    if (wide_category != NULL)
-                        tll_push_back(action->categories, wide_category);
+                if (locale_score > action->categories_locale_score) {
+                    for (const char *category = strtok_r(value, ";", &ctx);
+                         category != NULL;
+                         category = strtok_r(NULL, ";", &ctx))
+                    {
+                        char32_t *wide_category = ambstoc32(category);
+                        if (wide_category != NULL)
+                            tll_push_back(action->categories, wide_category);
+                    }
+
+                    action->categories_locale_score = locale_score;
                 }
             }
 
             else if (strcmp(key, "Actions") == 0) {
-                for (const char *action = strtok(value, ";");
+                for (const char *action = strtok_r(value, ";", &ctx);
                      action != NULL;
-                     action = strtok(NULL, ";"))
+                     action = strtok_r(NULL, ";", &ctx))
                 {
                     tll_push_back(action_names, strdup(action));
                 }
             }
 
+            else if (strcmp(key, "OnlyShowIn") == 0) {
+                for (const char *desktop = strtok_r(value, ";", &ctx);
+                     desktop != NULL;
+                     desktop = strtok_r(NULL, ";", &ctx))
+                {
+                    tll_push_back(action->onlyshowin, strdup(desktop));
+                }
+            }
+
+            else if (strcmp(key, "NotShowIn") == 0) {
+                for (const char *desktop = strtok_r(value, ";", &ctx);
+                     desktop != NULL;
+                     desktop = strtok_r(NULL, ";", &ctx))
+                {
+                    tll_push_back(action->notshowin, strdup(desktop));
+                }
+            }
+
             else if (strcmp(key, "Icon") == 0) {
                 free(action->icon);
                 action->icon = strdup(value);
@@ -215,7 +336,6 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                     action->use_terminal = true;
             }
         }
-
         free(line);
     }
 
@@ -230,6 +350,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
         {
             free(a->name);
             free(a->generic_name);
+            free(a->app_id);
             free(a->comment);
             free(a->icon);
             free(a->exec);
@@ -238,6 +359,8 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
 
             tll_free_and_free(a->keywords, free);
             tll_free_and_free(a->categories, free);
+            tll_free_and_free(a->onlyshowin, free);
+            tll_free_and_free(a->notshowin, free);
 
             continue;
         }
@@ -275,12 +398,13 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                 .basename = c32dup(file_basename),
                 .wexec = a->wexec,
                 .title = title,
+                .app_id = a->app_id,
                 .generic_name = a->generic_name,
                 .comment = a->comment,
                 .keywords = a->keywords,
                 .categories = a->categories,
                 .icon = {.name = a->icon},
-                .visible = a->visible,
+                .visible = a->visible && (!filter_desktops || filter_desktop_entry(a, desktops)),
                 .count = 0}));
     }
 
@@ -307,6 +431,7 @@ new_id(const char *base_id, const char *new_part)
 
 static void
 scan_dir(int base_fd, const char *terminal, bool include_actions,
+         bool filter_desktop, char_list_t *desktops,
          application_llist_t *applications, const char *base_id)
 {
     DIR *d = fdopendir(base_fd);
@@ -315,6 +440,59 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
         return;
     }
 
+    const char *locale = setlocale(LC_MESSAGES, NULL);
+    struct locale_variants lc_messages = {NULL};
+
+    if (locale != NULL) {
+        char *copy = strdup(locale);
+
+        char *lang = copy;
+        char *country_start = strchr(copy, '_');
+        char *encoding_start = strchr(copy, '.');
+        char *modifier_start = strchr(copy, '@');
+
+        if (country_start != NULL)
+            *country_start = '\0';
+        if (encoding_start != NULL)
+            *encoding_start = '\0';
+        if (modifier_start != NULL)
+            *modifier_start = '\0';
+
+        lc_messages.lang = copy;
+        if (country_start != NULL) {
+            if (asprintf(
+                    &lc_messages.lang_country, "%s_%s", lang, country_start + 1) < 0) {
+                LOG_WARN(
+                    "failed to construct lang_country string from %s and %s: %s",
+                    lang, country_start + 1, strerror(errno));
+            }
+        }
+
+        if (modifier_start != NULL) {
+            if (asprintf(
+                    &lc_messages.lang_modifier, "%s@%s", lang, modifier_start + 1) < 0) {
+                LOG_WARN(
+                    "failed to construct lang@modifier string from %s and %s: %s",
+                    lang, modifier_start + 1, strerror(errno));
+            }
+        }
+
+        if (country_start != NULL && modifier_start != NULL) {
+            if (asprintf(
+                    &lc_messages.lang_country_modifier,
+                    "%s_%s@%s", lang, country_start + 1, modifier_start + 1) < 0) {
+                LOG_WARN(
+                    "failed to construct lang_country@modifier string from %s, %s and %s: %s",
+                    lang, country_start + 1, modifier_start + 1, strerror(errno));
+            }
+        }
+
+        LOG_DBG("lang=%s, lang_country=%s, lang@modifier=%s, lang_country@modifier=%s",
+                lc_messages.lang, lc_messages.lang_country,
+                lc_messages.lang_modifier, lc_messages.lang_country_modifier);
+    }
+
+
     for (const struct dirent *e = readdir(d); e != NULL; e = readdir(d)) {
         if (strcmp(e->d_name, ".") == 0 || strcmp(e->d_name, "..") == 0)
             continue;
@@ -334,6 +512,7 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
 
             char *nested_base_id = new_id(base_id, e->d_name);
             scan_dir(dir_fd, terminal, include_actions,
+                     filter_desktop, desktops,
                      applications, nested_base_id);
             free(nested_base_id);
             close(dir_fd);
@@ -382,15 +561,19 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 if (!already_added) {
                     parse_desktop_file(
                         fd, id, wfile_basename, terminal, include_actions,
-                        applications);
+                        filter_desktop, desktops,
+                        &lc_messages, applications);
                 } else
                     free(id);
                 close(fd);
             }
         }
-
     }
 
+    free(lc_messages.lang);
+    free(lc_messages.lang_country);
+    free(lc_messages.lang_modifier);
+    free(lc_messages.lang_country_modifier);
     closedir(d);
 }
 
@@ -404,6 +587,7 @@ sort_application_by_title(const void *_a, const void *_b)
 
 void
 xdg_find_programs(const char *terminal, bool include_actions,
+                  bool filter_desktop, char_list_t *desktops,
                   struct application_list *applications)
 {
     application_llist_t apps = tll_init();
@@ -415,7 +599,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
 
         int fd = open(path, O_RDONLY);
         if (fd != -1) {
-            scan_dir(fd, terminal, include_actions, &apps, NULL);
+            scan_dir(fd, terminal, include_actions, filter_desktop, desktops, &apps, NULL);
             close(fd);
         }
     }
